import functools
import time

from llm_foundation import logger


def timer(fn):
    """Calculate the time to run a function"""
    @functools.wraps(fn)
    def decorator_timer(*args, **kwargs):
        start_time = time.perf_counter()
        return_value = fn(*args, **kwargs)
        end_time = time.perf_counter()
        runtime = end_time - start_time
        logger.info(f"Finished {fn.__name__}() in {runtime:.4f} secs")
        return return_value
    return decorator_timer


def show_banner(text, level=1, width=120):
    """
    Print text in a banner style with hierarchy representation.

    :param text: The text to be printed within the banner.
    :param level: Hierarchy level of the banner (1 for top-level).
    :param width: The width of the banner including the borders.
    """
    if len(text) + 4 > width:  # Ensure text fits in the banner
        width = len(text) + 4

    border_chars = ["*", "=", "-", "."]  # Different chars for different levels

    # Choose border character based on level, defaulting to the last item for higher levels
    border_char = border_chars[min(level - 1, len(border_chars) - 1)] * width

    padding = (width - len(text) - 2) // 2
    padded_text = " " * padding + text + " " * padding
    if len(text) % 2 != 0:  # Adjust for odd-length text
        padded_text += " "

    logger.info(border_char)
    logger.info(border_char[0] + padded_text + border_char[0])
    logger.info(border_char)


def banner(text, level=1, width=120, mark_fn_end=False):
    def decorator_banner(fn):
        @functools.wraps(fn)
        def wrapper_fn(*args, **kwargs):
           show_banner(text, level=level, width=width)
           return_value = fn(*args, **kwargs)
           if mark_fn_end:
               show_banner(f"End of {text}", level=level, width=width)
           return return_value
        return wrapper_fn 
    return decorator_banner



from enum import Enum
from typing import Optional

from pydantic import BaseModel
import os

class Provider(Enum):
    LC = 0
    # HF = 1
    OpenAI = 2
    Autogen = 3

class ModelType(Enum):
    Simple = 0
    Chat = 1

class LMConfig(BaseModel):
    model: str = "gpt4o-mini"
    provider: Provider = Provider.LC
    type: ModelType = ModelType.Chat
    temperature: float = 0.0
    max_tokens: int = 300

    def get_api_key(self) -> Optional[str]:
        match self.provider:
            case Provider.LC:
                return os.getenv("LC_API_KEY")
            case Provider.OpenAI:
                return os.getenv("OPENAI_API_KEY")
            case Provider.Autogen:
                return os.getenv("OPENAI_API_KEY")
            case _:
                raise ValueError("Invalid backend provider")

    def to_autogen(self, tools: Optional[list] = None):
                
        config_list = [{
            "model": self.model,  # model name
            "api_key": self.get_api_key()  # api key
        }]
        
        return {
            "seed": 14,  # seed for caching and reproducibility
            "functions": tools ,                
            "config_list": config_list,  # a list of API configurations
            "temperature": self.temperature,  # temperature for sampling
        }



from abc import ABC
from collections import defaultdict
from typing import List, Dict, Optional, Sequence, Union, get_origin

from langchain_core.tools import StructuredTool
from langchain_core.utils.function_calling import convert_to_openai_function
from langchain.output_parsers.openai_functions import JsonKeyOutputFunctionsParser, JsonOutputFunctionsParser, PydanticOutputFunctionsParser
from langchain.prompts import ChatPromptTemplate
from langchain.schema.runnable import RunnableLambda
from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain_community.document_loaders import WebBaseLoader
from langchain.pydantic_v1 import BaseModel, Field

from llm_foundation import logger


class Citation(BaseModel):
    """Information about papers mentioned in a text document."""
    title: str
    author: Optional[str]
    year: Optional[int]


class CitedDocuments(BaseModel):
    """Information to extract"""
    cites: List[Citation]


DEFAULT_WEB_EXTRACTOR_SYS_TEMPLATE = """A text document will be passed to you. Extract from it all papers \
that are mentioned/cited by the text. \
Do not extract the name of the article itself. If no papers are mentioned \
you don't need to extract any. Just return an empty list. \
Do not make up or guess ANY extra information. Only extract what exactly is in the text."""


class WebPageContentExtractor(ABC):
    
    def __init__(self, lm, entity_to_extract: BaseModel, sys_template: str):
        self.lm = lm
        self.entity_to_extract: BaseModel = entity_to_extract
        
        self.sys_template = sys_template
        
        # This extracts the first list attribute from the entity to extract (e.g. cites in CitedDocuments)
        # This will allow later to flatten the list of extractions
        self.first_list_attribute = None
        for name, field in self.entity_to_extract.model_fields.items():
            if get_origin(field.annotation) is list:
                self.first_list_attribute = name
                break
        
        # Preparing function call
        openai_fn_desc = convert_to_openai_function(self.entity_to_extract)
        extraction_function = [
            openai_fn_desc
        ]
        fn_call = {"name": openai_fn_desc["name"]}

        self.lm = self.lm.bind(
            functions=extraction_function,
            function_call=fn_call
        )

    def extract(self, source_web_docs: Union[str, Sequence[str]]):
        """Extracts information from a list of web documents."""
        
        def flatten(matrix):
            return [item for sublist in matrix for item in sublist]
                
        # Load documents
        loader = WebBaseLoader(source_web_docs)
        documents = loader.load()
        
        # TODO: Look for better splitters
        text_splitter = RecursiveCharacterTextSplitter(chunk_overlap=0)

        # Building chain
        prompt = ChatPromptTemplate.from_messages([
            ("system", self.sys_template),
            ("user", "{input}")
        ])

        parser_instance = JsonKeyOutputFunctionsParser(key_name=self.first_list_attribute) if self.first_list_attribute else JsonOutputFunctionsParser()
        extraction_chain = prompt | self.lm | parser_instance
        prep = RunnableLambda(
            lambda x: [{"input": doc} for doc in text_splitter.split_text(x)]
        )
        # First element in the chain has to be a Runnable lambda
        final_chain = prep | extraction_chain.map() | flatten
        
        # Extraction
        extracted_docs = {}
        logger.info(f"Extracting {self.entity_to_extract}")
        for i, document in enumerate(documents):
            logger.info(f"Document len: {len(document.page_content)}")
            # Remove empty lines from document.page_content
            content = "\n".join(line for line in document.page_content.split("\n") if line.strip())
            logger.info(f"Document len without empty lines: {len(content)}")
            doc_id = document.metadata.get("source", f"unknown_doc_{i}")            
            extracted_docs[doc_id] = final_chain.invoke(content)

        return extracted_docs

    @classmethod
    def build_tool(cls, lm, entity_to_extract: BaseModel, sys_template: str = DEFAULT_WEB_EXTRACTOR_SYS_TEMPLATE):
        extractor = WebPageContentExtractor(lm, entity_to_extract, sys_template)
        return StructuredTool.from_function(extractor.extract)


class Step(BaseModel):
    """Information about a plan step, referencing its parents step ids (dependent steps) and the children step ids (other steps this node depends on)."""
    id: str
    description: str
    dependent_steps: List[str] = Field(..., description="A list of step ids representing the actions that will happen after what is described this step. If the list is empty, the step is a root node.")
    depending_steps: List[str] = Field(..., description="A list of step ids representing the actions that need to happen before what is described this step. If the list is empty, the step is a leaf node.")
    
    def is_leaf(self):
        return not self.depending_steps
    
    def is_root(self):
        return not self.dependent_steps
    

class Plan(BaseModel):
    """Plan description depicted as a directed acyclic graph (DAG) where each step has a causal dependency on other steps."""
    plan: List[Step]

class MultiTreePlan(BaseModel):
    root_nodes: List[Step]
    
    def traverse_depth_first(self):
        """Traverses the root nodes of the MultiTreePlan in depth-first order."""
        visited = set()

        def dfs(node):
            visited.add(node.id)
            print(node.id)

            for child in node.depending_steps:
                if child.id not in visited:
                    dfs(child)

        for root in self.root_nodes:
            dfs(root)

    def traverse_breadth_first(self):
        """Traverses the root nodes of the MultiTreePlan in breadth-first order."""
        visited = set()
        queue = []

        for root in self.root_nodes:
            queue.append(root)

        while queue:
            node = queue.pop(0)
            visited.add(node.id)
            print(node.id)

            for child in node.depending_steps:
                if child.id not in visited:
                    queue.append(child)

    def traverse_leafs_first_dependants(self):
        """Traverses the root nodes of the MultiTreePlan in depth-first order."""
        visited = set()
        leaf_steps = []
        dependant_steps = []

        def dfs(node):
            visited.add(node.id)

            if not node.depending_steps:
                leaf_steps.append(node)
            else:
                for child in node.depending_steps:
                    if child.id not in visited:
                        dfs(child)
                dependant_steps.append(node)

        for root in self.root_nodes:
            dfs(root)

        return leaf_steps + dependant_steps
    
    def print_multi_tree(self):
        """Prints the multi-root node trees in a visually appealing format."""
        def print_node(node, indent):
            print(f"{indent}{node.id}: {node.description}")
            for child in node.depending_steps:
                print_node(child, indent + "  ")

        for root in self.root_nodes:
            print_node(root, "")
    

DEFAULT_PLAN_EXTRACTOR_SYS_TEMPLATE = """You are an expert in planning. When a text document depicting a plan is passed to you, \
you are very skilled at extracting the hierarchical plan steps from it. \
The hierarchycal structure can be described either in the form of a tree or a directed acyclic graph (DAG) \
Each step has a description and dependencies (parents and children) on other steps. \
First, assign a unique id to each step in the plan. \
Then think carefully about what steps need to happen before in the plan in order to assign the correct parents and children ids to each step. \
Return the list of steps that conform the plan. \
If no plan steps are mentioned or identified, just return an empty list. \
Do not make up or guess ANY extra information. Only extract the steps exactly as they are in the text."""


class PlanExtractor(ABC):
    
    def __init__(self, lm, plan: BaseModel, sys_template: str, use_pydantic_output: bool = False):
        self.lm = lm
        self.plan: BaseModel = plan
        self.sys_template = sys_template
        self.use_pydantic_output = use_pydantic_output
        
        # Preparing function call
        openai_fn_desc = convert_to_openai_function(self.plan)
        extraction_function = [
            openai_fn_desc
        ]
        fn_call = {"name": openai_fn_desc["name"]}

        self.lm = self.lm.bind(
            functions=extraction_function,
            function_call=fn_call
        )

    def extract(self, text: str):
        """Extracts from the text passed an itemized list of plan steps."""
        
        # Building chain
        prompt = ChatPromptTemplate.from_messages([
            ("system", self.sys_template),
            ("user", "Extract the hierarchical plan steps from this text: {input}")
        ])
        parser = PydanticOutputFunctionsParser(pydantic_schema=self.plan) if self.use_pydantic_output else JsonOutputFunctionsParser()
        extraction_chain = prompt | self.lm | parser
        
        # Extraction
        logger.info(f"Extracting plan steps")
        extracted_steps = extraction_chain.invoke(text)

        return extracted_steps

    @classmethod
    def build_tool(cls, lm, entity_to_extract: BaseModel, sys_template: str = DEFAULT_PLAN_EXTRACTOR_SYS_TEMPLATE, use_pydantic_output:bool = False):
        extractor = PlanExtractor(lm, entity_to_extract, sys_template, use_pydantic_output)
        return StructuredTool.from_function(extractor.extract)
    
    @classmethod
    def build_multi_tree_plan(cls, data: Union[dict, Plan]) -> MultiTreePlan:
        # Convert json object to Plan
        if isinstance(data, dict):
            data = Plan.parse_obj(data)
            
        id_to_step = {step.id: step for step in data.plan}
        child_to_parents = defaultdict(list)

        # Track parent-child relationships
        for step in data.plan:
            for dep in step.depending_steps:
                child_to_parents[dep].append(step.id)

        # Identify root nodes (steps without any parent dependencies)
        root_nodes = [step for step in data.plan if step.id not in child_to_parents]

        # Build the tree by linking steps to their children
        for step in data.plan:
            step.depending_steps = [id_to_step[dep] for dep in step.depending_steps]

        return MultiTreePlan(root_nodes=root_nodes) 



"""
llm_foundation module
"""
import logging
import os
from logging.config import fileConfig
from typing import List

import coloredlogs
from dotenv import load_dotenv, find_dotenv

# import sys
# import pkg_resources  # type: ignore
# if sys.version_info >= (3, 8):
#     from importlib import metadata
# else:
#     import importlib_metadata as metadata
# __version__: str = metadata.version(__name__) #"0.0.1"  # pkg_resources.get_distribution("llm_base").version


load_dotenv(find_dotenv()) # read local .env file and put it in os.environ

__all__: List[str] = []
__copyright__: str = "Copyright 2024, Francisco Perez-Sorrosal."

env_key = "LOGGER"
logger_name = os.getenv(env_key, "root")

env_key = "LOG_CFG"
config_path = os.getenv(env_key, None)

env_key = "LOG_DEST"
home_dir = os.getenv("HOME", "/var/log")
log_dir = os.getenv(env_key, home_dir)
env_key = "LOG_DEST_FILE"
log_filaname = os.getenv(env_key, "latte.log")
logfilename_path = os.path.join(log_dir, log_filaname)

env_key = "LOG_LEVEL"
log_level = os.getenv(env_key, "INFO")

print(f"Trying to configure logger {logger_name} in module {__name__}")
logger = logging.getLogger(logger_name)
print(f"{logger.name} # of associated handlers - {len(logger.handlers)}")

if len(logging.getLogger(logger_name).handlers) <= 0:
    print("Logging is not configured yet. Configuring it now.")

    if config_path is None or not os.path.exists(config_path):
        print("Basic logging config")
        logging.basicConfig(level=logging.INFO)
        logger = logging.getLogger()
    else:
        print(f"Loading logging config from file {config_path}. Output to: {logfilename_path}")
        if not os.path.exists(log_dir):
            print(f"Creating dir {log_dir} for placing the log as it does not exist yet")
            os.makedirs(log_dir)
        fileConfig(
            config_path,
            defaults={"logfilename": logfilename_path, "loglevel": log_level},
            disable_existing_loggers=False,
        )
        logger = logging.getLogger(logger_name)
    coloredlogs.install(logger=logger)
    logger.info(f"Logger {logger_name} configured")
else:
    logger.info(f"Logger {logger_name} already configured")



import json

from typing import Any, Callable, Dict, List, Literal, Optional, Union
from enum import Enum


from autogen.agentchat import ConversableAgent, GroupChat, AssistantAgent, UserProxyAgent, GroupChatManager
from pydantic import BaseModel
from pathlib import Path


from llm_foundation import logger


DEFAULT_TERMINATION_WORD="TERMINATE"

def is_termination_message(msg, termination_word: str = DEFAULT_TERMINATION_WORD) -> bool:
    '''Detects if we should terminate the conversation'''
    
    if isinstance(msg.get("content"), str):
        return msg["content"].rstrip().endswith(termination_word)
    elif isinstance(msg.get("content"), list):
        for content in msg["content"]:
            if isinstance(content, dict) and "text" in content:
                return content["text"].rstrip().endswith(termination_word)
    return False

def always_terminate(msg) -> bool:
    return True


def custom_speaker_selection_func(last_speaker, groupchat):
    """Define a customized speaker selection function.
    A recommended way is to define a transition for each speaker in the groupchat.

    Parameters:
        - last_speaker: Agent
            The last speaker in the group chat.
        - groupchat: GroupChat
            The GroupChat object
    Return:
        Return one of the following:
        1. an `Agent` class, it must be one of the agents in the group chat.
        2. a string from ['auto', 'manual', 'random', 'round_robin'] to select a default method to use.
        3. None, which indicates the chat should be terminated.
    """
    pass


class AutogenAgentType(Enum):
    ConversableAgent = 1
    UserProxyAgent = 2
    AssistantAgent = 3
    GroupChatManager = 4
    

class Role(BaseModel):
    description: str
    agent_system_message: str
    human_input_mode: Literal["ALWAYS", "NEVER", "TERMINATE"] = "TERMINATE"
    autogen_code_execution_config: dict = {}

    @classmethod
    def from_json_file(cls, file_path: str) -> 'Role':
        with open(file_path, 'r') as file:
            json_data = json.load(file)
        return cls(**json_data)
    
    def to_autogen_agent(self, 
                         name:str, 
                         type: AutogenAgentType, 
                         human_input_mode: Optional[Literal["ALWAYS", "NEVER", "TERMINATE"]] = None,
                         llm_config: Optional[dict] = None,
                         max_consecutive_auto_reply: Optional[int] = None,
                         group_chat: Optional[GroupChat] = None,
                         code_execution_config: Optional[Union[Dict, Literal[False]]] = False,
                         termination_function: Optional[Callable] = None) -> 'ConversableAgent':
        
        code_execution_config = code_execution_config if code_execution_config is not None else self.autogen_code_execution_config
        
        if termination_function is not None:
            termination_message_placeholder = lambda msg: termination_function(msg)
        else:
            termination_message_placeholder = None
        
        match type:
            case AutogenAgentType.ConversableAgent:
                return ConversableAgent(name=name, 
                                                system_message=self.agent_system_message,
                                                description=self.description,
                                                llm_config=llm_config,
                                                human_input_mode=human_input_mode if human_input_mode is not None else self.human_input_mode,
                                                max_consecutive_auto_reply=max_consecutive_auto_reply,
                                                is_termination_msg=termination_message_placeholder,
                                                code_execution_config=code_execution_config
                )
            case AutogenAgentType.AssistantAgent:
                return AssistantAgent(name=name, 
                                              system_message=self.agent_system_message, 
                                              description=self.description,
                                              llm_config=llm_config,
                                              human_input_mode=human_input_mode if human_input_mode is not None else self.human_input_mode,
                                              max_consecutive_auto_reply=max_consecutive_auto_reply,
                                              is_termination_msg=termination_message_placeholder,
                                              code_execution_config=code_execution_config,
                )
            case AutogenAgentType.UserProxyAgent:
                return UserProxyAgent(name=name, 
                                              system_message=self.agent_system_message, 
                                              description=self.description,
                                              llm_config=llm_config,
                                              human_input_mode=human_input_mode if human_input_mode is not None else self.human_input_mode,
                                              max_consecutive_auto_reply=max_consecutive_auto_reply,
                                              is_termination_msg=termination_message_placeholder,
                                              code_execution_config=code_execution_config,
                )
            case AutogenAgentType.GroupChatManager:
                if group_chat is None:
                    raise ValueError("Group chat is required for GroupChatManager")
                return GroupChatManager(name=name, 
                                                groupchat=group_chat,
                )
            case _:
                raise ValueError(f"Invalid agent type: {type}")


class Persona(BaseModel):
    name: str
    roles: Dict[str, Role]
    
    @classmethod
    def from_json_file(cls, file_path: str) -> 'Persona':
        with open(file_path, 'r') as file:
            json_data = json.load(file)
        return cls(**json_data)

    def save_as_json(self, 
                     path: Optional[Path] = None, 
                     file_name: Optional[str] = None, 
                     overwrite_existing: bool = True):
        if path is None:
            path = Path(self.__class__.__name__)
        if file_name is None:
            file_name = self.name
        if not file_name.endswith(".json"):
            file_name += ".json"
        path = path / file_name
        
        # Create the directory if it does not exist
        path.parent.mkdir(parents=True, exist_ok=overwrite_existing)
        
        with open(path, 'w') as file:
            file.write(self.model_dump_json(indent=4))
            
            
        
        logger.info(f"JSON object written: {path}")
        
    def get_roles(self) -> List[str]:
        return list(self.roles.keys())

    def role_to_autogen_agent(self,
                              role_name: str,
                              type: AutogenAgentType, 
                              human_input_mode:Literal["ALWAYS", "NEVER", "TERMINATE"] = "NEVER",
                              llm_config: Optional[dict] = None,
                              group_chat: Optional[GroupChat] = None,
                              code_execution_config: Optional[Union[Dict, Literal[False]]] = False,
                              termination_function: Optional[Callable] = None) -> 'ConversableAgent':
        
        return self.roles[role_name].to_autogen_agent(name=f"{role_name}", 
                                                       type=type, 
                                                       human_input_mode=human_input_mode,
                                                       llm_config=llm_config,
                                                       group_chat=group_chat,
                                                       code_execution_config=code_execution_config,
                                                       termination_function=termination_function)

    def __str__(self) -> str:
        def format_role(role_name: str, role: Role, indent: int) -> str:
            indent_str = " " * indent
            role_str = f"{indent_str}Role: {role_name}\n"
            role_str += f"{indent_str}Description: {role.description}\n"
            role_str += f"{indent_str}Agent System Message: {role.agent_system_message}\n"
            role_str += f"{indent_str}Autogen Code Execution Config: {role.autogen_code_execution_config}\n"
            return role_str

        def format_persona(persona: Persona, indent: int) -> str:
            indent_str = " " * indent
            persona_str = f"{indent_str}Persona: {persona.name}\n"
            for role_name, role in persona.roles.items():
                persona_str += format_role(role_name, role, indent + 2)
            return persona_str

        return format_persona(self, 0)
    
    
class Application(BaseModel):
    name: str
    roles: Dict[str, Role]
    
    @classmethod
    def from_json_file(cls, file_path: str) -> 'Application':
        with open(file_path, 'r') as file:
            json_data = json.load(file)
        return cls(**json_data)
    
    